import os, mimetypes, zipfile, tarfile
from datetime import datetime
from pkg_resources import resource_string, resource_exists, resource_isdir
from pasteob import *

__all__ = [
    'DataApp', 'FileApp', 'DirectoryApp',
    'ArchivedFilesApp', 'PkgResourcesApp'
]

class DataApp(Response):
    allowed_methods = ('GET', 'HEAD')
    default_conditional_response = True
    def __init__(self, *args, **kw):
        kw.setdefault('accept_ranges', 'bytes')
        super(DataApp, self).__init__(*args, **kw)

    def _set_body(self, body):
        Response.body.fset(self, body)
        self.last_modified = datetime.now()
        self.md5_etag(set_content_md5=True)

    body = property(Response.body.fget, _set_body)

    @wapp
    def __call__(self, req):
        if req.method not in self.allowed_methods:
            return HTTPMethodNotAllowed('You cannot %s a file' % req.method, allow=self.allowed_methods)
        return super(DataApp, self).__call__


CACHE_SIZE = 2**12

class FileApp(object):
    """
        An application that will send the file at the given filename.
        Adds a mime type based on ``mimetypes.guess_type()``.
    """

    def __init__(self, filename, **kw):
        self.filename = filename
        content_type, content_encoding = mimetypes.guess_type(filename)
        kw.setdefault('content_type', content_type)
        kw.setdefault('content_encoding', content_encoding)
        self.kw = kw
        self.last_modified = None

    def update(self, force=False):
        stat = os.stat(self.filename)
        if stat.st_mtime != self.last_modified or force:
            self.last_modified = stat.st_mtime
            if stat.st_size < CACHE_SIZE:
                data = open(self.filename, 'rb').read()
                self.cached_response = DataApp(data, last_modified=stat.st_mtime, **self.kw)
            else:
                self.cached_response = None
                self.content_length = stat.st_size

    @wapp
    def __call__(self, req):
        if req.method not in ('GET', 'HEAD'):
            return HTTPMethodNotAllowed('You cannot %s a file' % req.method, allow=('GET', 'HEAD'))
        force = (req.cache_control.max_age == 0)
        self.update(force) # RFC 2616 13.2.6
        r = self.cached_response # access just once to avoid the need for locking
        if r:
            return r
        elif not os.path.exists(self.filename):
            return HTTPNotFound(comment="No file at %r" % self.filename)

        try:
            file = open(self.filename, 'rb')
        except (IOError, OSError), e:
            return HTTPForbidden('You are not permitted to view this file (%s)' % e)
        return DataApp(
            app_iter = FileIter(file),
            content_length = self.content_length,
            last_modified = self.last_modified,
            #@@ etag
            **self.kw
        )


class FileIter(object):
    def __init__(self, file):
        self.file = file

    def app_iter_range(self, seek=None, limit=None, block_size=1<<16):
        if seek:
            self.file.seek(seek)
            if limit:
                limit -= seek
        try:
            while True:
                data = self.file.read(min(block_size, limit) if limit else block_size)
                if not data:
                    return
                yield data
                if limit:
                    limit -= len(data)
                    if not limit:
                        return
        finally:
            self.file.close()

    __iter__ = app_iter_range


class DirectoryApp(object):
    """
        An application that dispatches requests to corresponding FileApps based on PATH_INFO.
        FileApp instances are cached. This app double-checks not to serve any files that are not in a subdirectory.
        To customize FileApp instances creation override `make_fileapp` method.
    """

    def __init__(self, path):
        self.path = os.path.abspath(path)
        if not self.path.endswith(os.path.sep):
            self.path += os.path.sep
        assert os.path.isdir(self.path)
        self.cached_apps = {}

    make_fileapp = FileApp

    @wapp
    def __call__(self, req):
        path_info = req.path_info.lstrip('/')
        try:
            return self.cached_apps[path_info]
        except KeyError:
            path = os.path.abspath(os.path.join(self.path, path_info))
            if not path.startswith(self.path):
                return HTTPForbidden()
            elif os.path.isfile(path):
                return self.cached_apps.setdefault(path_info, self.make_fileapp(path))
            else:
                return HTTPNotFound(comment=path)


class ArchivedFilesApp(object):
    """
        An application that serves files from a zip or tar archive via DataApps.
    """
    def __init__(self, filepath, expires=None):
        if zipfile.is_zipfile(filepath):
            self.archive = zipfile.ZipFile(filepath, 'r')
        elif tarfile.is_tarfile(filepath):
            self.archive = tarfile.TarFileCompat(filepath, 'r')
        else:
            raise AssertionError("filepath '%s' is not a zip or tar " % filepath)
        self.expires = expires
        self.cache = {}

    @wapp
    def __call__(self, req):
        path = req.path_info.lstrip('/')
        if path not in self.cache:
            try:
                info = self.archive.getinfo(path)
            except KeyError:
                return
            if info.filename.endswith('/'):
                app = None
            else:
                content_type, content_encoding = mimetypes.guess_type(info.filename)
                app = DataApp(
                    body = self.archive.read(path),
                    content_type = content_type,
                    content_encoding = content_encoding,
                    last_modified = datetime(*info.date_time),
                    expires = self.expires,
                )
            self.cache[path] = app
        return self.cache[path]


class PkgResourcesApp(object):
    """
        An application to serve package resources as static files.
    """
    def __init__(self, module, prefix=''):
        self.module = module
        if prefix and not prefix.endswith('/'):
            prefix += '/'
        self.prefix = prefix
        self.cached_apps = {}

    @wapp
    def __call__(self, req):
        path_info = req.path_info.lstrip('/')
        try:
            return self.cached_apps[path_info]
        except KeyError:
            path = self.prefix + path_info
            if (not resource_exists(self.module, path)
                or resource_isdir(self.module, path)
            ):
                return HTTPNotFound()
            data = resource_string(self.module, path)
            content_type, content_encoding = mimetypes.guess_type(path_info)
            app = DataApp(data, content_type=content_type)
            app = self.cached_apps.setdefault(path_info, app)
            return app

