from pasteob.base import *
from warnings import warn

__all__ = ['fix_HEAD', 'override_https', 'override_errors', 'Cascade', 'SiteRoot']


@wapp_mw
def fix_HEAD(req, app):
    """
        HEAD requests must return exactly the same headers as a GET.
        Most apps / dispatchers / middleware do not honour this, so
        we have to cheat them.
    """
    if req.method == 'HEAD':
        req.method = 'GET'
        app = req.get_response(app)
        # disabled because env could be used in app_iter
        #req.method = 'HEAD'
    return app


@wapp_mw
def override_https(req, normal_app, secure_app):
    """
        Return a WSGI app that dispatches HTTP requests to normal_app and HTTPS to secure_app.
        Use override_https('redirect', secure_app) to force HTTPS connections.
    """
    if normal_app == 'redirect' and req.scheme == 'http':
        return HTTPSeeOther(location='https' + req.url[4:])
    return secure_app if req.scheme == 'https' else normal_app


#print Request.blank('/').get_response(override_https(None, None))


@wapp_mw
def override_errors(req, app, overrides={}):
   r = req.get_response(app)
   return overrides.get(r.status_int, r)




class Cascade(object):
    class __metaclass__(type):
        def __call__(cls, apps, *args, **kw):
            if len(apps) == 0:
                raise ValueError("Need at least one app for the cascade")
            elif len(apps) == 1:
                return apps[0]
            else:
                return type.__call__(cls, apps, *args, **kw)

    def __init__(self, apps, catch=(404,), exhaust_app_iters=False):
        for item in catch:
            assert isinstance(item, int)
        self.apps = apps
        self.catch = frozenset(catch)
        self.exhaust_app_iters = exhaust_app_iters

    @wapp
    def __call__(self, req):
        for app in self.apps[:-1]:
            resp = req.copy().get_response(app)
            if resp.status_int not in self.catch:
                return resp
            elif self.exhaust_app_iters:
                for item in resp.app_iter:
                    pass
        else:
            return req.get_response(self.apps[-1])



class SiteRoot(object):
    overrides = {
        '/robots.txt': 'robots',
        '/favicon.ico': 'favicon',
    }

    def __init__(self, app):
        self.app = app
        self.robots = Response(content_type='text/plain')
        self.overrides = self.overrides.copy()

    @wapp
    def __call__(self, req):
        if req.script_name:
            warn("SiteRoot should be mounted at site root")
        if req.path_info in self.overrides:
            app = self.overrides[req.path_info]
            if isinstance(app, str):
                app = getattr(self, app, None)
            return app
        return self.app
